// Copyright (C) FIT VUT 
// Petr Lampa <lampa@fit.vutbr.cz>
// $Id$
// vi:set ts=8 sts=8 sw=8:

// RFC791
class IPv4Address
{
private:
	Octet _addr[4];
public:
	IPv4Address() { _addr[0] = _addr[1] = _addr[2] = _addr[3] = 0; }
	explicit IPv4Address(const string &str) { (void)addr(str); }
	// default copy constructor ok
	int length() const { return 4; }
	bool addr(const string &str);
	const Octet *octet_addr() const { return _addr; }
	inline string addr() const { return to_string(); }
	string to_string(int level = 0) const;
	// rfc4291 
	inline bool is_unspecified() const {	// ::
		for (int i = 0; i < 4; i++) {
			if (_addr[i]) return false;
		}
		return true;
	}
	inline bool is_loopback() const {	// ::1
		if (_addr[0] != 127) return false;
		return true;
	}
	inline bool is_linklocal() const {
		if (_addr[0] != 169 || _addr[1] != 254) return false;
		return true;
	}
	inline bool is_private() const {	// 10, 178, 196
		if (_addr[0] != 10 &&
		    (_addr[0] != 172 || (_addr[1] & 0xf0) != 0x10) &&
		    (_addr[0] != 192 || _addr[1] != 168)) return false;
		return true;
	}
	inline bool is_multicast() const { // ff00::-
		if ((_addr[0] & 0xf0) != 0xe0) return false;
		if (_addr[0] != 255 || _addr[1] != 255 || _addr[2] != 255 || _addr[3] != 255) return false;
		return true;
	}
	inline bool is_prefix(const IPv4Address &pfx, int len) const {
		int i=0;
		if (len > 32) return false;
		while (len > 8) {
			if (_addr[i] != pfx._addr[i]) return false;
			len -= 8; i++;
		}
		if (len == 0) return true;
		Octet mask = 0xffu << len;
		if ((_addr[i] & mask) != (pfx._addr[i] & mask)) return false;
		return true;
	}

	inline bool operator==(const IPv4Address &adr) const {
		for (int i = 0; i < 4; i++) {
			if (_addr[i] != adr._addr[i]) return false;
		}
		return true;
	}
	inline bool operator!=(const IPv4Address &adr) const { return !(*this ==(adr)); }
	inline unsigned checksum() const { 
		return (_addr[0]<<8) + _addr[1] + (_addr[2]<<8) + _addr[3];
	}
	inline bool build(Buffer &pkt, int phase) {
		for (int i = 0; i < 4; i++) pkt.add_octet(_addr[i]);
		return true;
	}
	inline bool decode(Buffer &pkt) {
		for (int i = 0; i < 4; i++) {
			if (!pkt.get_octet(_addr[i])) return false;
		}
		return true;
	}
	~IPv4Address() { }
};

inline ostream &operator<<(ostream &out, const IPv4Address &adr)
{
	return out << adr.to_string();
}

#ifdef _LIB
#include <sys/socket.h>
#include <arpa/inet.h>

bool IPv4Address::addr(const string &str)
{
	if (inet_pton(AF_INET, str.c_str(), _addr) == 1) return true;
	return false;
}

string IPv4Address::to_string(int level) const
{
	char str[INET_ADDRSTRLEN];
	if (inet_ntop(AF_INET, _addr, str, INET_ADDRSTRLEN)) return string(str);
	return "0.0.0.0";
}
#endif

// +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
// |Version|  IHL  |    DSCP   |ECN|          Total Length         |
// +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
// |         Identification        |Flags|      Fragment Offset    |
// +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
// |  Time to Live |    Protocol   |         Header Checksum       |
// +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
// |                       Source Address                          |
// +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
// |                    Destination Address                        |
// +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
// |                    Options                    |    Padding    |
// +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
// RFC791 3.1
// RFC3168 23.1
class IPv4: public Packet 
{
private:
	Octet _version;	// IP version number = 4
	Octet _hlen;	// header length in 4 octet units
	Octet _tos;	// type of service (DSCP + ECN)
	Short _plen;	// payload length including header
	Short _id;	// identification
	Short _foff;	// fragment offset
	Octet _ttl;	// time to live
	Octet _proto;	// type of header immediately following the IPv4 header
	Short _checksum;	// header checksum
	IPv4Address _src;	// source address
	IPv4Address _dest;	// destination address
	Word _options;		// options
	Octet *_frag;	// fragment data
	bool _own;	
	IPv4(const IPv4&);
	IPv4& operator=(const IPv4&);
public:

	IPv4(): _version(4), _hlen(5), _tos(0), _plen(0), _id(0), _foff(0), _ttl(64), _proto(0), _checksum(0), _src(), _dest(), _options(0), _frag(0), _own(false) { }
	explicit IPv4(const string &dest): _version(4), _hlen(5), _tos(0), _plen(0), _id(0), _foff(0), _ttl(64), _proto(0), _checksum(0), _src(), _dest(dest), _options(0), _frag(0), _own(false) { }
	explicit IPv4(const string &src, const string &dest): _version(4), _hlen(5), _tos(0), _plen(0), _id(0), _foff(0), _ttl(64), _proto(0), _checksum(0), _src(src), _dest(dest), _options(0), _frag(0), _own(false) { }

	inline bool is_fragment() const { return (_foff & 0x1fff) != 0; }
	inline bool is_df() const { return (_foff & 0x4000); }
	inline bool is_mf() const { return (_foff & 0x2000); }
	inline void set_df() { _foff |= 0x4000; }
	inline void set_mf() { _foff |= 0x2000; }
	inline Octet *fragment_data() { return _frag; }
	inline void fragment_data(Octet *data, Short len) {
		_plen = len+20;
		_frag = data;
		_own = false;
	}
	inline Short fragment_length() const { return _plen - _hlen*4; }
	inline Short fragment_offset() const { return (_foff&0x1fff)*8; }
	inline bool fragment_offset(Short off) { 
		if (off > 65535-20) return false;
		if (off & 0x7) return false;
		_foff = (_foff & 0xe000) | ((off>>3) & 0x1fff); 
		return true;
	}
	inline bool dest(const string &str) { return _dest.addr(str); }
	inline const IPv4Address& dest() const { return _dest; }
	inline bool src(const string &str) { return _src.addr(str); }
	inline const IPv4Address& src() const { return _src; }
	inline bool dscp(unsigned val) { 
		if (val > 63) return false;
		_tos = val<<2; return true;
	}
	inline int dscp() const { return _tos>>2; }
	inline bool ident(unsigned id) {
		if (id >= 1<<16) return false;
		_id = id; return true;
	}
	inline int ident() const { return _id; }
	inline bool plen(unsigned len) { 
		if (len > 65535) return false;
		_plen = len; return true;
	}
	inline int plen() const { return _plen; }
	inline bool proto(unsigned proto) { 
		if (proto > 255) return false;
		_proto = proto; return true;
	}
	inline int proto() const { return _proto; }
	inline bool ttl(unsigned ttl) {
		if (ttl > 255) return false;
		_ttl = ttl; return true;
	}
	inline int ttl() const { return _ttl; }

	string to_string(int level = 0) const;
	string name() const { return "IPv4"; }

	Word pseudo_checksum(Octet proto, Word plen) const;

	void fixup();
	bool do_build(Buffer &pkt, int phase, Word &pos);
	bool decode(Buffer &pkt);

	~IPv4() { if (_frag && _own) delete _frag; }

        inline static void add_proto(Octet code, Packet *(p)(Buffer &pkt)) { _map_p[code] = p; }

private:
	inline static Packet *proto_factory(Octet code, Buffer &pkt) {
		if (_map_p[code]) return (*_map_p[code])(pkt);
		return NULL;
	}
	static Packet *(*_map_p[256])(Buffer &pkt);
};

#ifdef _LIB
static Packet *IPv4_factory()
{
	Packet *p = new IPv4();
	// if (debug > 1) cerr << "new IPv4 instance " << p << endl;
	return p;
}

class autoinit
{
public:
	autoinit() { Ethernet::add_proto(0x0800, IPv4_factory);  }
};
static autoinit init;
// proto table
Packet *(*IPv4::_map_p[256])(Buffer &pkt);

string IPv4::to_string(int level) const
{
	string ret("<IPv4");
	if (_version != 4) {
		ret += " version=";
		ret += cvt_int(_version);
	}
	if (_hlen != 5) {
		ret += " hlen=";
		ret += cvt_int(_hlen);
	}
	if (_tos != 0) {
		ret += " dscp=";
		ret += cvt_hex(_tos>>2);
		ret += " ecn=";	//non-ECT,ECT(1),ECT(0),CE
		ret += cvt_hex(_tos&3);
	}
	ret += " plen=";
	ret += cvt_int(_plen);
	if (_id) {
		ret += " id=";
		ret += cvt_int(_id);
	}
	if (is_df()) ret += " DF";
	if (is_mf()) ret += " MF";
	if (is_fragment()) {
		ret += " foff=";
		ret += cvt_int(fragment_offset());
	}
	ret += " ttl=";
	ret += cvt_int(_ttl);
	ret += " proto=";
	ret += cvt_int(_proto);
	ret += " chk=";
	ret += cvt_hex(_proto);
	ret += " src=";
	ret += _src.to_string(level);
	ret += " dst=";
	ret += _dest.to_string(level);
//	for (int i = 0; i < _num_exts; i++) {
//		ret += " | ";
//		ret += _exts[i]->to_string(level);
//	}
	ret += ">";
	return ret;
}

void IPv4::fixup()
{
	Ethernet *e;
	if (underlayer() &&     // if in ethernet packet, setup type
	    (e = dynamic_cast<Ethernet *>(underlayer())) != NULL) e->type(0x800);
}

// +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
// +                         Source Address                        +
// +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
// +                      Destination Address                      +
// +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
// |    zero       |    protocol   |         payload length        |
// +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
Word IPv4::pseudo_checksum(Octet proto, Word plen) const
{
	// TODO - fix plen
	return _src.checksum() + _dest.checksum() + (plen&0xffff) + proto;
}
	
bool IPv4::do_build(Buffer &pkt, int phase, Word &pos)
{
	Word checksum;
	// get the extension headers size, insert required paddings
	if (phase == 0) {
		pos += 20;
//		for (int i = 0; i < _num_exts; i++) {
//			if (!_exts[i]->build(pkt, phase, pos)) return false;
//		}
		if (payload()) {
			if (!payload()->do_build(pkt, phase, pos)) return false;
		}
		_plen = pos;	// all extension headers + payload
		checksum = ((_version&0xf) << 12) + ((_hlen & 0xf) << 8) + _tos; 
		checksum += _plen;
		checksum += _id;
		checksum += _foff;
		checksum += (_ttl<<8) + _proto;
		checksum += _src.checksum();
		checksum += _dest.checksum();
		checksum = (checksum & 0xffff) + (checksum >> 16);
		if (checksum > 65535) checksum -= 65535;
		_checksum = ~checksum;
		// TODO - if _plen > MTU, perform fragmentation
		return true;
	}
	if (!pkt.add_octet(((_version&0xf)<<4)+(_hlen&0xf))) return false;
	if (!pkt.add_octet(_tos)) return false;
	if (!pkt.add_hshort(_plen)) return false;
	if (!pkt.add_hshort(_id)) return false;
	if (!pkt.add_hshort(_foff)) return false;
	if (!pkt.add_octet(_ttl)) return false;
	if (!pkt.add_octet(_proto)) return false;
	if (!pkt.add_hshort(_checksum)) return false;
	if (!_src.build(pkt, phase)) return false;
	if (!_dest.build(pkt, phase)) return false;
	pos += 20;
//	for (int i = 0; i < _num_exts; i++) {
//		if (!_exts[i]->build(pkt, phase, pos)) return false;
//	}
	if (payload()) {	// build upper layer packet
		if (!payload()->do_build(pkt, phase, pos)) return false;
	}
	return true;
}

bool IPv4::decode(Buffer &pkt)
{
	Octet b1;
	if (_frag && _own) delete _frag;
	_own = false;
	_frag = 0;
	release();		// release previous payload
	pkt.set();
	if (!pkt.get_octet(b1) || !pkt.get_octet(_tos) || !pkt.get_hshort(_plen)) {
		pkt.reset();
		return false;
	}
	_version = b1 >> 4;				// 28..31
	if (_version != 4) {
		if (debug > 0) cerr << "IPV4::decode(): invalid IPv4 version " << (int)_version << endl;
		return false;
	}
	_hlen = (b1 & 0xf);
	if (!pkt.get_hshort(_id) || !pkt.get_hshort(_foff)) {
		pkt.reset();
		return false;
	}
	if (!pkt.get_octet(_ttl) || !pkt.get_octet(_proto) || !pkt.get_hshort(_checksum)) {
		pkt.reset();
		return false;
	}
	if (!_src.decode(pkt) || !_dest.decode(pkt)) {
		pkt.reset();
		return false;
	}
	// TODO - process options
	if (_hlen > 5) {
		int l = _hlen*4-20;
		for (int i = 0; i < l; i++) {
			if (!pkt.get_octet(b1)) return false;
		}
	}
	if (_foff & 0x1fff) {	// fragment offset
		int l = _plen - _hlen*4;
		_frag = new Octet[l];
		_own = true;
		for (int i = 0; i < l; i++) {
			if (!pkt.get_octet(_frag[i])) return false;
		}
		return true;
	}
	// decode payload
	Packet *p = proto_factory(_proto, pkt);
	if (p) {
		attach(p, true);
		if (!p->decode(pkt)) return false;
	}
	return true;
}
#endif
